@@ -193,7 +193,7 @@ check_time (char const *file_name, struct timespec t)
   if (t.tv_sec <= 0)
     WARN ((0, 0, _("%s: implausibly old time stamp %s"),
 	   file_name, tartime (t, true)));
-  else if (timespec_cmp (start_time, t) < 0)
+  else if (timespec_cmp (volume_start_time, t) < 0)
     {
       struct timespec now;
       gettime (&now);
@@ -392,13 +392,12 @@ static int
 make_directories (char *file_name)
 {
   char *cursor0 = file_name + FILE_SYSTEM_PREFIX_LEN (file_name);
-  char *cursor;			/* points into the file name */
+  char *cursor;	        	/* points into the file name */
   int did_something = 0;	/* did we do anything yet? */
   int mode;
   int invert_permissions;
   int status;
 
-
   for (cursor = cursor0; *cursor; cursor++)
     {
       if (! ISSLASH (*cursor))
@@ -428,7 +427,7 @@ make_directories (char *file_name)
 	     invert_permissions is zero, because
 	     repair_delayed_set_stat may need to update the struct.  */
 	  delay_set_stat (file_name,
-			  &current_stat_info /* ignored */,
+			  &current_stat_info,
 			  invert_permissions, INTERDIR_PERMSTATUS);
 
 	  print_for_mkdir (file_name, cursor - file_name, mode);
@@ -647,7 +646,7 @@ extract_dir (char *file_name, int typeflag)
 		}
 	      if (S_ISDIR (st.st_mode))
 		{
-		  mode = st.st_mode & ~ current_umask;
+		  mode = st.st_mode;
 		  break;
 		}
 	    }
@@ -668,12 +667,16 @@ extract_dir (char *file_name, int typeflag)
   if (status == 0
       || old_files_option == DEFAULT_OLD_FILES
       || old_files_option == OVERWRITE_OLD_FILES)
-    delay_set_stat (file_name, &current_stat_info,
-		    MODE_RWX & (mode ^ current_stat_info.stat.st_mode),
-		    (status == 0
-		     ? ARCHIVED_PERMSTATUS
-		     : UNKNOWN_PERMSTATUS));
-
+    {
+      if (status == 0)
+	delay_set_stat (file_name, &current_stat_info,
+			MODE_RWX & (mode ^ current_stat_info.stat.st_mode),
+			ARCHIVED_PERMSTATUS);
+      else /* For an already existing directory, invert_perms must be 0 */
+	delay_set_stat (file_name, &current_stat_info,
+			0,
+			UNKNOWN_PERMSTATUS);
+    }
   return status;
 }
 
@@ -1027,6 +1030,13 @@ extract_mangle_wrapper (char *file_name, int typeflag)
   return 0;
 }
 
+static int
+extract_volhdr (char *file_name, int typeflag)
+{
+  if (verbose_option)
+    fprintf (stdlis, _("Reading %s\n"), quote (current_stat_info.file_name));
+  skip_member ();
+}
 
 static int
 extract_failure (char *file_name, int typeflag)
@@ -1107,9 +1117,7 @@ prepare_to_extract (char const *file_name, int typeflag, tar_extractor_t *fun)
       break;
 
     case GNUTYPE_VOLHDR:
-      if (verbose_option)
-	fprintf (stdlis, _("Reading %s\n"), quote (current_stat_info.file_name));
-      *fun = NULL;
+      *fun = extract_volhdr;
       break;
 
     case GNUTYPE_NAMES:
@@ -1174,51 +1182,36 @@ void
 extract_archive (void)
 {
   char typeflag;
-  char *file_name;
   tar_extractor_t fun;
 
   set_next_block_after (current_header);
   decode_header (current_header, &current_stat_info, &current_format, 1);
-
-  if (interactive_option && !confirm ("extract", current_stat_info.file_name))
+  if (!current_stat_info.file_name[0]
+      || (interactive_option
+	  && !confirm ("extract", current_stat_info.file_name)))
     {
       skip_member ();
       return;
     }
 
   /* Print the block from current_header and current_stat.  */
-
   if (verbose_option)
     print_header (&current_stat_info, -1);
 
-  file_name = safer_name_suffix (current_stat_info.file_name,
-                                 false, absolute_names_option);
-  if (strip_name_components)
-    {
-      size_t prefix_len = stripped_prefix_len (file_name,
-					       strip_name_components);
-      if (prefix_len == (size_t) -1)
-	{
-	  skip_member ();
-	  return;
-	}
-      file_name += prefix_len;
-    }
-
   /* Restore stats for all non-ancestor directories, unless
      it is an incremental archive.
      (see NOTICE in the comment to delay_set_stat above) */
   if (!delay_directory_restore_option)
-    apply_nonancestor_delayed_set_stat (file_name, 0);
+    apply_nonancestor_delayed_set_stat (current_stat_info.file_name, 0);
 
   /* Take a safety backup of a previously existing file.  */
 
   if (backup_option)
-    if (!maybe_backup_file (file_name, 0))
+    if (!maybe_backup_file (current_stat_info.file_name, 0))
       {
 	int e = errno;
 	ERROR ((0, e, _("%s: Was unable to backup this file"),
-		quotearg_colon (file_name)));
+		quotearg_colon (current_stat_info.file_name)));
 	skip_member ();
 	return;
       }
@@ -1228,9 +1221,10 @@ extract_archive (void)
   typeflag = sparse_member_p (&current_stat_info) ?
                   GNUTYPE_SPARSE : current_header->header.typeflag;
 
-  if (prepare_to_extract (file_name, typeflag, &fun))
+  if (prepare_to_extract (current_stat_info.file_name, typeflag, &fun))
     {
-      if (fun && (*fun) (file_name, typeflag) && backup_option)
+      if (fun && (*fun) (current_stat_info.file_name, typeflag)
+	  && backup_option)
 	undo_last_backup ();
     }
   else
@@ -1319,6 +1313,39 @@ extract_finish (void)
   apply_nonancestor_delayed_set_stat ("", 1);
 }
 
+bool
+rename_directory (char *src, char *dst)
+{
+  if (rename (src, dst))
+    {
+      int e = errno;
+
+      switch (e)
+	{
+	case ENOENT:
+	  if (make_directories (dst))
+	    {
+	      if (rename (src, dst) == 0)
+		return true;
+	      e = errno;
+	    }
+	  break;
+		    
+	case EXDEV:
+	  /* FIXME: Fall back to recursive copying */
+	  
+	default:
+	  break;
+	}
+
+      ERROR ((0, e, _("Cannot rename %s to %s"),
+	      quote_n (0, src),
+	      quote_n (1, dst)));
+      return false;
+    }
+  return true;
+}
+      
 void
 fatal_exit (void)
 {
